// -*- C++ -*-
/*!
 * @file  BioSim01.cpp * @brief BioSim controller * $Date$ 
 *
 * $Id$ 
 */
#include <iostream>
#include "BioSim01.h"

namespace { const bool CONTROLLER_BRIDGE_DEBUG = true; }

// Module specification
// <rtc-template block="module_spec">
static const char* biosim01_spec[] =
  {
    "implementation_id", "BioSim01",
    "type_name",         "BioSim01",
    "description",       "BioSim component",
    "version",           "0.1",
    "vendor",            "AIST",
    "category",          "Generic",
    "activity_type",     "SPORATIC",
    "kind",              "DataFlowComponent",
    "max_instance",      "1",
    "language",          "C++",
    "lang_type",         "compile",
    // Configuration variables
    ""
  };
// </rtc-template>

BioSim01::BioSim01(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_angleIn("angle", m_angle),
    m_torqueOut("torque", m_torque),
    // </rtc-template>
    dummy(0)
{
}

BioSim01::~BioSim01()
{
  if(gstream.is_open()) gstream.close();
  delete [] arrive;
  delete [] gain;
  delete [] goal;
  delete [] cold;
  delete [] pold;
  delete [] speed;
  delete [] state;
  delete [] start;
  delete [] steps;
  delete [] target;
  delete [] track;
}


RTC::ReturnCode_t BioSim01::onInitialize()
{
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort("angle", m_angleIn);

  // Set OutPort buffer
  addOutPort("torque", m_torqueOut);

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable

  arrive = new int[SERVO_NUM];
  gain = new int[SERVO_NUM];
  goal = new int[SERVO_NUM];
  speed = new int[SERVO_NUM];
  state = new int[SERVO_NUM];
  start = new int[SERVO_NUM];
  steps = new int[SERVO_NUM];
  target = new int[SERVO_NUM];
  track = new int[SERVO_NUM];
  torque = new int[SERVO_NUM];

  pold = new double[DOF];
  cold = new double[DOF];

  // Open joint position file
  gstream.open(GOAL_FILE);
  if(!gstream.is_open())
  {
    std::cerr << GOAL_FILE << " not found " << std::endl;
  }

  // Open proportional gain file
  std::ifstream gfile;
  gfile.open(GAIN_FILE);
  if(gfile.is_open())
  {
    int g;
    for(int i=0; i<SERVO_NUM; i++)
    {
      gfile >> g;
      gain[i] = 2 << g;
    }
    gfile.close();
  }
  else
  {
    std::cerr << GAIN_FILE << " not found " << std::endl;
    for(int i=0; i<SERVO_NUM; i++) gain[i] = 32;
  }

  // Ensure angle and torque data length
  m_angle.data.length(DOF);
  m_torque.data.length(DOF);

  // Initialize state array
  for(int i=0; i<SERVO_NUM; i++) state[i] = MODE_START;

  // Initialize speed array
  for(int i=0; i<SERVO_NUM; i++) speed[i] = 0;

  // Initialize tracking array
  for(int i=0; i<SERVO_NUM; i++) track[i] = 0;

  // Initialize torque array
  for(int i=0; i<SERVO_NUM; i++) torque[i] = 0;

  // </rtc-template>
  return RTC::RTC_OK;
}


/*
RTC::ReturnCode_t BioSim01::onFinalize()
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onStartup(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

RTC::ReturnCode_t BioSim01::onActivated(RTC::UniqueId ec_id)
{
  std::cout << "onActivated" << std::endl;

  // Read first goal file record
  gstream.seekg(0);
  gstream >> time;
  for(int i=0; i<SERVO_NUM; i++) gstream >> goal[i];
  gstream >> next_sync;

  // Read present joint angle data
  if(m_angleIn.isNew())
  {
    m_angleIn.read();
  }

  // Preserve values of the previous frame
  for(int i=0; i<DOF; i++)
  {
    pold[i] = m_angle.data[i];
    cold[i] = 0;
  }

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t BioSim01::onDeactivated(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

// Helper functions
int angle2pos(double angle)
{
  if(angle < -2.61799) return 0;
  if(angle > 2.61799) return(1023);
  // Conversion factor = (511 * 6) / (5 * PI)
  return(511 + (angle * 195.1876222));
}

double pos2angle(int pos)
{
  if(pos > 1023) return(2.61799);
  if(pos < 0) return(-2.61799);
  // Conversion factor = (5 * PI) / (6 * 511)
  return((pos - 511) * 0.005113269);   
}

RTC::ReturnCode_t BioSim01::onExecute(RTC::UniqueId ec_id)
{
  int	distance;
  int	power;
  int	togo;

  double delta;
  double omega;
  double volts;
  double current;
  double limit;
  double friction;
 
  // Read inport if necessary
  if(m_angleIn.isNew())
  {
    m_angleIn.read();
  }

  // Update simulation time
  time += TIMESTEP;

  // if(CONTROLLER_BRIDGE_DEBUG) std::cout << "onExecute:" << time << std::endl;

  // Read new goals of necessary
  if(time > next_sync)
  {
    for(int i=0; i<SERVO_NUM; i++)
    {
      gstream >> goal[i];
      state[i] = MODE_START;
    }
    gstream >> next_sync;

    // if(CONTROLLER_BRIDGE_DEBUG) std::cout << "Goals read" << std::endl;
  }

  // Duplicate dynamixel torque (PWM) generation code
  for(int i=0; i<DOF; i++)
  {
     switch(state[i])
    {
      case MODE_START:

        if(speed[i] == 0)
        {
          target[i] = goal[i];
          state[i] = MODE_TRACK;
          arrive[i] = 0;
          break; 
        }
        else
        {
          double goal_angle = pos2angle(goal[i]);
          if(((goal_angle < m_angle.data[i]) && (speed[i] >= 1.0)) ||
             ((goal_angle >= m_angle.data[i]) && (speed[i] < 0.0)))
          {
            speed[i] = - speed[i];
          }
          start[i] = target[i];
          state[i] = MODE_SPEED;
          steps[i] = 0;
        }
 
      case MODE_SPEED:
        track[i]++;
        if((track[i] & 0x07) == 0)
        {
          steps[i]++;
          target[i] = start[i] + (((speed[i] * steps[i]) + 64) / 128);
          double goal_angle = pos2angle(goal[i]);
          if((((start[i] > goal_angle) || (m_angle.data[i] > goal_angle)) && (speed[i] >= 0.0)) ||
             (((start[i] < goal_angle) || (m_angle.data[i] < goal_angle)) && (speed[i] < 0.0)))
          {
            start[i] = goal[i];
            state[i] = MODE_TRACK;
          }
        }
        break;
    }
    
    distance = target[i] - angle2pos(m_angle.data[i]);

    if(CWRange1 < distance)
    {
      if(CWRange3 < distance)
      {
        togo = distance;
        power = CWPower3;
      }
      else
      {
        if(CWRange2 < distance)
        {
          togo = distance;
          power = CWPower2;
        }
        else
        {
          togo = (CWMargin == 0x0400) ? distance : distance - CWRange1;
          power = CWPower1;
        }
      }

      torque[i] = (togo << power) + Punch;

      if(CWMargin < distance) torque[i] += Punch;

      if(torque[i] < 0)
      {
        if(torque[i] < -Torque_Limit) torque[i] = -Torque_Limit;
      }
      else
      {
        if(torque[i] > Torque_Limit) torque[i] = Torque_Limit;
      }
    }
    else
    {
      if(distance < -CCWRange1)
      {
        if(distance < -CCWRange3)
        {
          togo = -distance;
          power = CCWPower3;
        }
        else
        {
          if(distance < -CCWRange2)
          {
           togo = -distance;
           power = CCWPower2;
          }
          else
          {
            togo = (CCWMargin == 0x0400) ? -distance : -distance - CCWRange1;
            power = CCWPower1;
          }
        }
       
        torque[i] = (togo << power) + Punch;

        if(distance < -CCWMargin) torque[i] += Punch;

        torque[i] = -torque[i];

        if(torque[i] < 0)
        {
          if(torque[i] < -Torque_Limit) torque[i] = -Torque_Limit;
        }
        else
        {
          if(torque[i] > Torque_Limit) torque[i] = Torque_Limit;
        }
      }
      else
      {
        torque[i] = 0;
      }
    }

    // omega = simulated angular velocity: angular distance / time increment
    omega = (m_angle.data[i] - pold[i]) / TIMESTEP;
    pold[i] = m_angle.data[i];

    // delta = desired angular velocity: torque * 59 rpm / 1023 (Torque_Limit) * 2 * PI / 60
    delta = torque[i] * 0.00604;

    // volts = desired voltage: proportional distance (PWM) * constant
    volts = torque[i] * 12.0 / Torque_Limit;

    // current = current estimate: (volts - back EMF) / resistance
    current = (volts - omega * 0.006810) / 8.0;

    // apply current limit: (dI/dt * inductance / resistance)
    current -= (current - cold[i]) * 0.005 / (TIMESTEP * 8.0);
    cold[i] = current;

    // friction = gear friction: velocity * velocity constant + sign(velocity) * force constant 
    friction = (omega < 0) ? omega * 0.575 - 0.00077 : omega * 0.575 + 0.00077;

    // if(CONTROLLER_BRIDGE_DEBUG) std::cout << current << ":" << friction << std::endl;

    // m_torque.data[i] = simulation torque: current * constant - friction
    m_torque.data[i] = current - friction;

  }

  // Output torque values
  m_torqueOut.write();

  return RTC::RTC_OK;
}

/*
RTC::ReturnCode_t BioSim01::onAborting(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onError(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onReset(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t BioSim01::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/

extern "C"
{
 
  void BioSim01Init(RTC::Manager* manager)
  {
    coil::Properties profile(biosim01_spec);
    manager->registerFactory(profile,
                             RTC::Create<BioSim01>,
                             RTC::Delete<BioSim01>);
  }
  
};



